@@ -1,4 +1,5 @@
 # core
+import pandas as pd
 import logging, warnings
 import itertools as it
 import os, psutil
@@ -13,16 +14,16 @@ from sklearn.metrics import get_scorer
 from sklearn.metrics import make_scorer
 
 
-# timeout functions
-from func_timeout import func_timeout, FunctionTimedOut
-
 # configspace
-import ConfigSpace
 from ConfigSpace.util import *
 from ConfigSpace.read_and_write import json as config_json
 import json
+import traceback
 
-def get_class( kls ):
+import pynisher
+
+
+def get_class(kls):
     parts = kls.split('.')
     module = ".".join(parts[:-1])
     m = __import__( module )
@@ -30,34 +31,57 @@ def get_class( kls ):
         m = getattr(m, comp)            
     return m
 
+
 def is_component_defined_in_steps(steps, name):
     candidates = [s[1] for s in steps if s[0] == name]
     return len(candidates) > 0
 
+
 def get_step_with_name(steps, name):
     candidates = [s for s in steps if s[0] == name]
     return candidates[0]
 
+
 def get_scoring_name(scoring):
-    return scoring if type(scoring) == str else scoring["name"]
+    return scoring if isinstance(scoring, str) else scoring["name"]
+
+
+def build_scorer(scoring):
+    return get_scorer(scoring) if isinstance(scoring, str) else make_scorer(
+        **{key: val for key, val in scoring.items() if key != "name"})
 
 
 class EvaluationPool:
 
-    def __init__(self, X, y, scoring, evaluation_fun = None, side_scores = None, tolerance_tuning = 0.05, tolerance_estimation_error = 0.01, logger_name = None):
+    def __init__(self,
+                 task_type,
+                 X,
+                 y,
+                 scoring,
+                 evaluation_fun=None,
+                 side_scores=None,
+                 tolerance_tuning=0.05,
+                 tolerance_estimation_error=0.01,
+                 logger_name=None,
+                 use_caching=True
+                 ):
+        domains_task_type = ["classification", "regression"]
+        if task_type not in domains_task_type:
+            raise ValueError(f"task_type must be in {domains_task_type}")
+        self.task_type = task_type
         self.logger = logging.getLogger('naiveautoml.evalpool' if logger_name is None else logger_name)
-        if X is None:
-            raise Exception("Parameter X must not be None")
+
+        if not isinstance(X, (pd.DataFrame, np.ndarray, scipy.sparse.csr.csr_matrix, scipy.sparse.csc.csc_matrix, scipy.sparse.lil.lil_matrix)):
+            raise TypeError(f"X must be a numpy array but is {type(X)}")
         if y is None:
-            raise Exception("Parameter y must not be None")
-        if type(X) != np.ndarray and type(X) != scipy.sparse.csr.csr_matrix and type(X) != scipy.sparse.csc.csc_matrix and type(X) != scipy.sparse.lil.lil_matrix:
-            raise Exception("X must be a numpy array but is " + str(type(X)))
+            raise TypeError("Parameter y must not be None")
+
         self.X = X
         self.y = y
         self.scoring = scoring
         if side_scores is None:
             self.side_scores = []
-        elif type(side_scores) == list:
+        elif isinstance(side_scores, list):
             self.side_scores = side_scores
         else:
             self.logger.warning("side scores was not given as list, casting it to a list of size 1 implicitly.")
@@ -67,39 +91,52 @@ class EvaluationPool:
         self.tolerance_estimation_error = tolerance_estimation_error
         self.cache = {}
         self.evaluation_fun = self.cross_validate if evaluation_fun is None else evaluation_fun
+        self.use_caching = use_caching
 
     def tellEvaluation(self, pl, scores, timestamp):
         spl = str(pl)
-        self.cache[spl] = (pl, scores, timestamp)
+        self.cache[spl] = (spl, scores, timestamp)
         score = np.mean(scores)
         if score > self.bestScore:
             self.bestScore = score        
             self.best_spl = spl
                         
-    def cross_validate(self, pl, X, y, scorings, errors="message"): # just a wrapper to ease parallelism
-        warnings.filterwarnings('ignore', module = 'sklearn')
+    def cross_validate(self, pl, X, y, scorings, errors="raise"): # just a wrapper to ease parallelism
+        warnings.filterwarnings('ignore', module='sklearn')
+        warnings.filterwarnings('ignore', module='numpy')
         try:
-            if type(scorings) != list:
+            if not isinstance(scorings, list):
                 scorings = [scorings]
-            skf = sklearn.model_selection.StratifiedKFold(n_splits=5, random_state=None, shuffle=True)
+
+            if self.task_type == "classification":
+                splitter = sklearn.model_selection.StratifiedKFold(n_splits=5, random_state=None, shuffle=True)
+            elif self.task_type:
+                splitter = sklearn.model_selection.KFold(n_splits=5, random_state=None, shuffle=True)
             scores = {get_scoring_name(scoring): [] for scoring in scorings}
-            for train_index, test_index in skf.split(X, y):
-                
-                X_test = X[test_index]
-                y_test = y[test_index]
+            for train_index, test_index in splitter.split(X, y):
                 
+                X_train = X.iloc[train_index] if isinstance(X, pd.DataFrame) else X[train_index]
+                y_train = y.iloc[train_index] if isinstance(y, pd.Series) else y[train_index]
+                X_test = X.iloc[test_index] if isinstance(X, pd.DataFrame) else X[test_index]
+                y_test = y.iloc[test_index] if isinstance(y, pd.Series) else y[test_index]
+
                 pl_copy = sklearn.base.clone(pl)
-                pl_copy.fit(X[train_index], y[train_index])
+                pl_copy.fit(X_train, y_train)
                 
                 # compute values for each metric
                 for scoring in scorings:
-                    scorer = get_scorer(scoring) if type(scoring) == str else make_scorer(**{key: val for key, val in scoring.items() if key != "name"})
+                    scorer = build_scorer(scoring)
                     try:
-                        score = scorer(pl_copy, X[test_index], y[test_index])
+                        score = scorer(pl_copy, X_test, y_test)
                     except KeyboardInterrupt:
                         raise
                     except:
-                        score = np.nan
+                        if errors in ["message", "ignore"]:
+                            if errors == "message":
+                                self.logger.info(f"Observed exception in validation of pipeline {pl_copy}. Placing nan.")
+                            score = np.nan
+                        else:
+                            raise
                         
                     scores[get_scoring_name(scoring)].append(score)
             return scores
@@ -113,40 +150,34 @@ class EvaluationPool:
             else:
                 raise
 
-    def evaluate(self, pl, timeout=None, deadline=None):
+    def evaluate(self, pl, timeout=None):
         if is_pipeline_forbidden(pl):
             self.logger.info(f"Preventing evaluation of forbidden pipeline {pl}")
             return {get_scoring_name(scoring): np.nan for scoring in [self.scoring] + self.side_scores}
-        
+
         process = psutil.Process(os.getpid())
         self.logger.info(f"Initializing evaluation of {pl} with current memory consumption {int(process.memory_info().rss / 1024 / 1024)} MB. Now awaiting results.")
         
         start_outer = time.time()
         spl = str(pl)
-        if spl in self.cache:
+        if self.use_caching and spl in self.cache:
             out = {get_scoring_name(scoring): np.nan for scoring in [self.scoring] + self.side_scores}
             out[get_scoring_name(self.scoring)] = np.round(np.mean(self.cache[spl][1]), 4)
             return out
         timestamp = time.time()
         if timeout is not None:
-            scores = func_timeout(timeout, self.evaluation_fun, (pl, self.X, self.y, [self.scoring] + self.side_scores))
+            with pynisher.limit(self.evaluation_fun, wall_time=timeout, context="fork") as limited_evaluation:
+                scores = limited_evaluation(pl, self.X, self.y, [self.scoring] + self.side_scores)
         else:
             scores = self.evaluation_fun(pl, self.X, self.y, [self.scoring] + self.side_scores)
         if scores is None:
             return {get_scoring_name(scoring): np.nan for scoring in [self.scoring] + self.side_scores}
         runtime = time.time() - start_outer
-        if type(scores) != dict:
-            raise ValueError(f"scores is of type {type(scores)} but must be a dictionary with entries for {get_scoring_name(self.scoring)}.\nProbably you inserted an evaluation_fun argument that does not return a proper dictionary.")
+        if not isinstance(scores, dict):
+            raise TypeError(f"scores is of type {type(scores)} but must be a dictionary with entries for {get_scoring_name(self.scoring)}.\nProbably you inserted an evaluation_fun argument that does not return a proper dictionary.")
         self.logger.info(f"Completed evaluation of {spl} after {runtime}s. Scores are {scores}")
         self.tellEvaluation(pl, scores[get_scoring_name(self.scoring)], timestamp)
         return {scoring: np.round(np.mean(scores[scoring]), 4) for scoring in scores}
-
-    def getBestCandidate(self):
-        return self.getBestCandidates(1)[0]
-        
-    def getBestCandidates(self, n):
-        candidates = sorted([key for key in self.cache], key=lambda k: np.mean(self.cache[k][1]), reverse=True)
-        return [self.cache[c] for c in candidates[:n]]
     
 def fullname(o):
     module = o.__module__
@@ -675,7 +706,7 @@ def compile_pipeline_by_class_and_params(clazz, params, X, y):
     if clazz == sklearn.discriminant_analysis.LinearDiscriminantAnalysis:
         if params["shrinkage"] in (None, "none", "None"):
             shrinkage_ = None
-            solver = 'svd'
+            solver = 'lsqr'
         elif params["shrinkage"] == "auto":
             shrinkage_ = 'auto'
             solver = 'lsqr'
@@ -683,7 +714,7 @@ def compile_pipeline_by_class_and_params(clazz, params, X, y):
             shrinkage_ = float(params["shrinkage_factor"])
             solver = 'lsqr'
         else:
-            raise ValueError(self.shrinkage)
+            raise ValueError()
 
         tol = float(params["tol"])
         return sklearn.discriminant_analysis.LinearDiscriminantAnalysis(shrinkage=shrinkage_, tol=tol, solver=solver)
@@ -938,17 +969,17 @@ def compile_pipeline_by_class_and_params(clazz, params, X, y):
     
     
 def get_hyperparameter_space_size(config_space):
-    hps = config_space.get_hyperparameters();
+    hps = config_space.get_hyperparameters()
     if not hps:
         return 0
     size = 1
     for hp in hps:
-        if type(hp) in [ConfigSpace.hyperparameters.UnParametrizedHyperparameter, ConfigSpace.hyperparameters.Constant]:
+        if isinstance(hp, (ConfigSpace.hyperparameters.UnParametrizedHyperparameter, ConfigSpace.hyperparameters.Constant)):
             continue
             
-        if type(hp) == ConfigSpace.hyperparameters.CategoricalHyperparameter:
+        if isinstance(hp, ConfigSpace.hyperparameters.CategoricalHyperparameter):
             size *= len(list(hp.choices))
-        elif issubclass(hp.__class__, ConfigSpace.hyperparameters.IntegerHyperparameter):
+        elif isinstance(hp, ConfigSpace.hyperparameters.IntegerHyperparameter):
             size *= (hp.upper - hp.lower + 1)
         else:
             return np.inf
@@ -960,15 +991,15 @@ def get_all_configurations(config_space):
     domains = []
     for hp in config_space.get_hyperparameters():
         names.append(hp.name)
-        if type(hp) in [ConfigSpace.hyperparameters.UnParametrizedHyperparameter, ConfigSpace.hyperparameters.Constant]:
+        if isinstance(hp, (ConfigSpace.hyperparameters.UnParametrizedHyperparameter, ConfigSpace.hyperparameters.Constant)):
             domains.append([hp.value])
             
-        if type(hp) == ConfigSpace.hyperparameters.CategoricalHyperparameter:
+        if isinstance(hp, ConfigSpace.hyperparameters.CategoricalHyperparameter):
             domains.append(list(hp.choices))
-        elif issubclass(hp.__class__, ConfigSpace.hyperparameters.IntegerHyperparameter):
+        elif isinstance(hp, ConfigSpace.hyperparameters.IntegerHyperparameter):
             domains.append(list(range(hp.lower, hp.upper + 1)))
         else:
-            raise Exception("Unsupported type " + str(type(hp)))
+            raise TypeError(f"Unsupported hyperparameter type {type(hp)}")
     
     # compute product
     configs = []
@@ -1005,7 +1036,8 @@ sklearn.preprocessing.PowerTransformer, "classifier": sklearn.naive_bayes.Multin
 
 class HPOProcess:
     
-    def __init__(self, step_name, comp, X, y, scoring, side_scores, evaluation_fun, execution_timeout, mandatory_pre_processing, other_step_component_instances, index_in_steps, max_time_without_imp, max_its_without_imp, min_its = 10, logger_name = None, allow_exhaustive_search = True):
+    def __init__(self, task_type, step_name, comp, X, y, scoring, side_scores, evaluation_fun, execution_timeout, mandatory_pre_processing, other_step_component_instances, index_in_steps, max_time_without_imp, max_its_without_imp, min_its = 10, logger_name = None, allow_exhaustive_search = True):
+        self.task_type = task_type
         self.step_name = step_name
         self.index_in_steps = index_in_steps
         self.comp = comp
@@ -1028,13 +1060,13 @@ class HPOProcess:
         self.max_its_without_imp = max_its_without_imp
         self.min_its = min_its
         self.scoring = scoring
-        self.pool = EvaluationPool(X, y, scoring = scoring, side_scores = side_scores, evaluation_fun = evaluation_fun)
+        self.pool = EvaluationPool(task_type, X, y, scoring=scoring, side_scores = side_scores, evaluation_fun = evaluation_fun)
         self.its = 0
         self.allow_exhaustive_search = allow_exhaustive_search
         
         if side_scores is None:
             self.side_scores = []
-        elif type(side_scores) == list:
+        elif isinstance(side_scores, list):
             self.side_scores = side_scores
         else:
             self.logger.warning("side scores was not given as list, casting it to a list of size 1 implicitly.")
@@ -1052,10 +1084,15 @@ class HPOProcess:
         
     def evalComp(self, params):
         try:
-            return self.pool.evaluate(self.get_parametrized_pipeline(params), timeout=self.execution_timeout)
-        except FunctionTimedOut:
+            return "ok", self.pool.evaluate(self.get_parametrized_pipeline(params), timeout=self.execution_timeout), None
+        except pynisher.WallTimeoutException:
             self.logger.info("TIMEOUT")
-            return {scoring: np.nan for scoring in [self.scoring] + self.side_scores}
+            return "timeout", {scoring: np.nan for scoring in [self.scoring] + self.side_scores}, None
+        except KeyboardInterrupt:
+            raise
+        except:
+            return "exception", {scoring: np.nan for scoring in [self.scoring] + self.side_scores}, traceback.format_exc()
+
         
     def step(self, remaining_time = None):
         self.its += 1
@@ -1072,9 +1109,9 @@ class HPOProcess:
 
         # evaluate configured pipeline
         time_start_eval = time.time()
-        scores = self.evalComp(params)
-        if type(scores) != dict:
-            raise ValueError(f"The scores must be a dictionary as a function of the scoring functions. Observed type is {type(scores)}: {scores}")
+        status, scores, exception = self.evalComp(params)
+        if not isinstance(scores, dict):
+            raise TypeError(f"The scores must be a dictionary as a function of the scoring functions. Observed type is {type(scores)}: {scores}")
         score = scores[get_scoring_name(self.scoring)]
         runtime = time.time() - time_start_eval
         self.eval_runtimes.append(runtime)
@@ -1102,7 +1139,7 @@ class HPOProcess:
                 configs = get_all_configurations(self.config_space)
                 self.logger.info(f"Now evaluation all {len(configs)} possible configurations.")
                 for params in configs:
-                    scores = self.evalComp(params)
+                    status, scores, exception = self.evalComp(params)
                     score = scores[self.scoring]
                     self.logger.info(f"Observed score of {score} for {self.comp['class']} with params {params}")
                     if score > self.best_score:
@@ -1110,7 +1147,7 @@ class HPOProcess:
                         self.best_score = score
                         self.best_params = params
                 self.logger.info("Configuration space completely exhausted.")
-        return self.get_parametrized_pipeline(params), scores, runtime
+        return self.get_parametrized_pipeline(params), status, scores, runtime, exception
     
     def get_best_config(self):
         return self.best_params
\ No newline at end of file
