@@ -8,6 +8,7 @@ from sklearn.compose import ColumnTransformer
 
 # naiveautoml commons
 from naiveautoml.commons import *
+import numpy as np
 import traceback
 
 
@@ -29,9 +30,37 @@ class NaiveAutoML:
                  logger_name=None,
                  show_progress=False,
                  opt_ordering=None,
-                 strictly_naive=False,
-                 sparse=False,
-                 task_type="auto"):
+                 strictly_naive: bool = False,
+                 sparse: bool = None,
+                 task_type: str = "auto",
+                 raise_errors: bool = False):
+        """
+
+        :param search_space:
+        :param scoring:
+        :param side_scores:
+        :param evaluation_fun:
+        :param num_cpus:
+        :param execution_timeout:
+        :param max_hpo_iterations:
+        :param max_hpo_iterations_without_imp:
+        :param max_hpo_time_without_imp:
+        :param timeout:
+        :param standard_classifier:
+        :param standard_regressor:
+        :param logger_name:
+        :param show_progress:
+        :param opt_ordering:
+        :param strictly_naive:
+        :param sparse: whether data is treated sparsely in pre-processing.
+                Default is `None`, and in that case, the sparsity is inferred automatically.
+                It is tested whether the mandatory pipeline can execute `fit_transform` without a numpy memory exception.
+                If so, no sparsity is applied.
+                Otherwise, it is assumed that not enough memory is available, and sparsity is enabled.
+
+        :param task_type:
+        :param raise_errors:
+        """
 
         self.search_space = search_space
         if isinstance(search_space, str):
@@ -57,6 +86,7 @@ class NaiveAutoML:
         self.stage_entrypoints = {}
         self.standard_classifier = standard_classifier
         self.standard_regressor = standard_regressor
+        self.raise_errors = raise_errors
 
         # state variables
         self.start_time = None
@@ -66,7 +96,7 @@ class NaiveAutoML:
         self.chosen_attributes = None
         
         # mandatory pre-processing steps
-        self.sparse = sparse # do one-hot encoding via sparse representations (default is False since this is not supported by all algorithms)
+        self.sparse = sparse
         self.mandatory_pre_processing = None
         
         self.task_type = task_type
@@ -84,7 +114,25 @@ class NaiveAutoML:
         """
         # infer task type
         if self.task_type == "auto":
-            return "regression" if len(np.unique(y)) > 100 else "classification"
+            if type(self.scoring) == str:
+                return "regression" if self.scoring in [
+                    "explained_variance",
+                    "max_error",
+                    "neg_mean_absolute_error",
+                    "neg_mean_squared_error",
+                    "neg_root_mean_squared_error",
+                    "neg_mean_squared_log_error",
+                    "neg_median_absolute_error",
+                    "r2",
+                    "neg_mean_poisson_deviance",
+                    "neg_mean_gamma_deviance",
+                    "neg_mean_absolute_percentage_error",
+                    "d2_absolute_error_score",
+                    "d2_pinball_score",
+                    "d2_tweedie_score"
+                ] else "classification"
+            else:
+                return "regression" if len(np.unique(y)) > 100 else "classification"
         else:
             return self.task_type
 
@@ -254,7 +302,10 @@ class NaiveAutoML:
                 timeout = False
                 status = "ok"
                 try:
-                    timeout = min(self.execution_timeout, remaining_time if self.deadline is not None else 10**10)
+                    if self.execution_timeout is None:
+                        timeout = None
+                    else:
+                        timeout = min(self.execution_timeout, remaining_time if self.deadline is not None else 10**10)
                     scores = pool.evaluate(pl, timeout)
                 except KeyboardInterrupt:
                     raise
@@ -265,6 +316,8 @@ class NaiveAutoML:
                 except:
                     exception = traceback.format_exc()
                     status = "exception"
+                    if self.raise_errors:
+                        raise
                 if status != "ok":
                     scores = {scoring: np.nan for scoring in [self.scoring] + (self.side_scores if self.side_scores is not None else [])}
                 score = scores[get_scoring_name(self.scoring)]
@@ -299,6 +352,7 @@ class NaiveAutoML:
             if decision is None:
                 if step_name == "learner":
                     self.logger.error("No learner was chosen in the initial phase. This is typically caused by too low timeouts or bugs in a custom scoring function (if applicable).")
+                    break
                 self.logger.debug("No component chosen for this slot. Leaving it blank")
             else:
                 self.logger.debug(f"Added {decision['class']} as the decision for step {step_name}")
@@ -428,6 +482,65 @@ class NaiveAutoML:
         if self.show_progress:
             pbar.close()
 
+    def get_mandatory_preprocessing(self, X, y, categorical_features):
+
+        # determine categorical attributes and necessity of binarization
+        sparse_training_data = isinstance(X, (scipy.sparse.csr.csr_matrix, scipy.sparse.lil.lil_matrix))
+        if isinstance(X, pd.DataFrame):
+            if categorical_features is None:
+                categorical_features = list(X.select_dtypes(exclude=np.number).columns)
+            else:
+                categorical_features = [c if not isinstance(c, int) else X.columns[c] for c in categorical_features]
+            numeric_features = [c for c in X.columns if c not in categorical_features]
+
+        elif isinstance(X, np.ndarray) or sparse_training_data:
+            if categorical_features is None:
+                types = [set([type(v) for v in r]) for r in X.T]
+                categorical_features = [c for c, t in enumerate(types) if len(t) != 1 or list(t)[0] == str]
+            numeric_features = [c for c in range(X.shape[1]) if c not in categorical_features]
+        else:
+            raise TypeError(
+                f"Given data X is of type {type(X)} but must be pandas dataframe, numpy array or sparse scipy matrix.")
+
+        # check necessity of imputation
+        missing_values_per_feature = np.sum(pd.isnull(X), axis=0)
+        self.logger.info(f"There are {len(categorical_features)} categorical features, which will be binarized.")
+        self.logger.info(f"Missing values for the different attributes are {missing_values_per_feature}.")
+        numeric_transformer = Pipeline([("imputer", sklearn.impute.SimpleImputer(strategy="median"))])
+
+        # get preprocessing steps
+        try_dense = self.sparse in [False, None]
+        steps = self.get_preprocessing_steps(categorical_features, missing_values_per_feature, numeric_transformer, numeric_features, not try_dense)
+        if not steps or self.sparse is not None:
+            return steps
+        try:
+            Pipeline(steps).fit_transform(X, y)
+            return steps
+        except np.core._exceptions._ArrayMemoryError:
+            return self.get_preprocessing_steps(categorical_features, missing_values_per_feature, numeric_transformer,
+                                                 numeric_features, sparse=True)
+        except:
+            raise
+
+    @staticmethod
+    def get_preprocessing_steps(categorical_features, missing_values_per_feature, numeric_transformer, numeric_features, sparse):
+        if type(sparse) != bool:
+            raise ValueError(f"`sparse` must be a bool but is {type(sparse)}")
+        if len(categorical_features) > 0 or sum(missing_values_per_feature) > 0:
+            categorical_transformer = Pipeline([
+                ("imputer", sklearn.impute.SimpleImputer(strategy="most_frequent")),
+                ("binarizer", sklearn.preprocessing.OneHotEncoder(handle_unknown='ignore', sparse_output=sparse)),
+
+            ])
+            return [("impute_and_binarize", ColumnTransformer(
+                transformers=[
+                    ("num", numeric_transformer, numeric_features),
+                    ("cat", categorical_transformer, categorical_features),
+                ]
+            ))]
+        else:
+            return []
+
     def reset(self, X, y, categorical_features=None):
         # register search space
         self.register_search_space(X, y)
@@ -451,43 +564,7 @@ class NaiveAutoML:
                 Timeout per execution: {self.execution_timeout}
                 Scoring: {self.scoring}""")
 
-        # determine categorical attributes and necessity of binarization
-        self.sparse_training_data = isinstance(X, (scipy.sparse.csr.csr_matrix, scipy.sparse.lil.lil_matrix))
-        if isinstance(X, pd.DataFrame):
-            if categorical_features is None:
-                categorical_features = list(X.select_dtypes(exclude=np.number).columns)
-            else:
-                categorical_features = [c if not isinstance(c, int) else X.columns[c] for c in categorical_features]
-            numeric_features = [c for c in X.columns if not c in categorical_features]
-
-        elif isinstance(X, np.ndarray) or self.sparse_training_data:
-            if categorical_features is None:
-                types = [set([type(v) for v in r]) for r in X.T]
-                categorical_features = [c for c, t in enumerate(types) if len(t) != 1 or list(t)[0] == str]
-            numeric_features = [c for c in range(X.shape[1]) if not c in categorical_features]
-        else:
-            raise TypeError(
-                f"Given data X is of type {type(X)} but must be pandas dataframe, numpy array or sparse scipy matrix.")
-
-        # check necessity of imputation
-        missing_values_per_feature = np.sum(pd.isnull(X), axis=0)
-        self.logger.info(f"There are {len(categorical_features)} categorical features, which will be binarized.")
-        self.logger.info(f"Missing values for the different attributes are {missing_values_per_feature}.")
-        numeric_transformer = Pipeline([("imputer", sklearn.impute.SimpleImputer(strategy="median"))])
-        if len(categorical_features) > 0 or sum(missing_values_per_feature) > 0:
-            categorical_transformer = Pipeline([
-                ("imputer", sklearn.impute.SimpleImputer(strategy="most_frequent")),
-                ("binarizer", sklearn.preprocessing.OneHotEncoder(handle_unknown='ignore', sparse=self.sparse)),
-
-            ])
-            self.mandatory_pre_processing = [("impute_and_binarize", ColumnTransformer(
-                transformers=[
-                    ("num", numeric_transformer, numeric_features),
-                    ("cat", categorical_transformer, categorical_features),
-                ]
-            ))]
-        else:
-            self.mandatory_pre_processing = []
+        self.mandatory_pre_processing = self.get_mandatory_preprocessing(X, y, categorical_features)
 
         # print overview
         summary = ""
@@ -503,29 +580,35 @@ class NaiveAutoML:
         
         # choose algorithms
         self.choose_algorithms(X, y)
-        self.steps_after_which_algorithm_selection_was_completed = len(self.history)
-        
-        # tune parameters
-        self.tune_parameters(X, y)
+        if self.decisions:
+            self.steps_after_which_algorithm_selection_was_completed = len(self.history)
 
-        # train final pipeline
-        self.logger.info("--------------------------------------------------")
-        self.logger.info("Search Completed. Building final pipeline.")
-        self.logger.info("--------------------------------------------------")
-        self.pl = self.build_pipeline(self.hpo_processes, X, y)
-        self.logger.info(self.pl)
-        self.logger.info("Now fitting the pipeline with all given data.")
-        while True:
-            try:
-                self.pl.fit(X, y)
-                break
-            except:
-                self.logger.warning("There was a problem in building the pipeline, cutting it one down!")
-                self.pl = Pipeline(steps=self.pl.steps[1:])
-                self.logger.warning("new pipeline is:", self.pl)
+            # tune parameters
+            self.tune_parameters(X, y)
 
+            # train final pipeline
+            self.logger.info("--------------------------------------------------")
+            self.logger.info("Search Completed. Building final pipeline.")
+            self.logger.info("--------------------------------------------------")
+            self.pl = self.build_pipeline(self.hpo_processes, X, y)
+            self.logger.info(self.pl)
+            self.logger.info("Now fitting the pipeline with all given data.")
+            while len(self.pl.steps) > 0:
+                try:
+                    self.pl.fit(X, y)
+                    break
+                except:
+                    self.logger.warning("There was a problem in building the pipeline, cutting it one down!")
+                    self.pl = Pipeline(steps=self.pl.steps[1:])
+                    self.logger.warning("new pipeline is:", self.pl)
+
+            if len(self.pl.steps) > 0:
+                self.chosen_model = self.pl
+            else:
+                self.logger.warning("For some reason (should happen), the final pipeline was reduced to empty.")
+        else:
+            self.logger.info("No model was chosen in first phase, so there is nothing to return for me ...")
         self.end_time = time.time()
-        self.chosen_model = self.pl
 
         # compile history
         history_keys = ["time"]
